#! /usr/bin/env python

from collections import defaultdict
import random
import socket
from tempfile import mkdtemp
import time

from contracts import contract
from irc.bot import SingleServerIRCBot
from irc.client import ServerNotConnectedError, MessageTooLong

from . import __version__
from .crypto import CryptoKey
from .dcache import  create_propose_message
from .dcache import Envelope, CouldNotReadEnvelope
from .fragments import fragmentize, Assembler
from .ipfs_utils import IPFSInterface


class SwarmBot(SingleServerIRCBot):

    def __init__(self, servers, stream_name, mi, channel, nickname):
        SingleServerIRCBot.__init__(self, servers, nickname, nickname)
        self.mi = mi
        self.target = channel

        self.stream_name = stream_name
        self.seen = {}
        self.tmpdir = mkdtemp(prefix='swarm')

        self.admitted = False
        self.disconnected = False
        self.last_broadcast = 0

        self.lateness = random.uniform(0, 10)
        self.nusers = 10

        self.nreceived = 0
        self.assemblers = defaultdict(Assembler)

    def on_pubmsg(self, c, e):
        self._handle(c, e)

    def on_privmsg(self, c, e):
        self._handle(c, e)

    def _handle(self, _c, e):
        self.nreceived += 1
        s = str(e.arguments[0])
        # {'source': u'andrea!~andrea@nutonomy02.n.subnet.rcn.com',
        # 'tags': [], 'type': 'pubmsg', 'target': u'#duckiebots', 'arguments': [u'deplep']}
        source = e.source
        if '!' in source:
            source = source[:source.index('!')]

        sn = '/' + str(source)

        self.assemblers[sn].push(s)
        msgs = self.assemblers[sn].pop()
        for msg in msgs:
            try:
                envelope = Envelope.from_json(msg)
            except CouldNotReadEnvelope as e:
                em = 'Could not read envelope from %s: %s' % (sn, e)
                print(em)
            else:
                envelope.maddr_from = sn + envelope.maddr_from
                self.mi.dispatch(envelope)

    def on_welcome(self, connection, _event):
#        print('welcome message from %s: \n%s' % (connection.server, " ".join(event.arguments)))
        connection.join(self.target)
        print('Connected to %s on IRC server %s.' % (self.target, connection.server))
        self.admitted = True

    def on_join(self, _c, _e):
        users = self.channels[self.target].users()
        self.nusers = len(users)
#        print('current users: %s' % users)

    def on_disconnect(self, _c, event):
        print('Disconnected', event.__dict__)
        self.disconnected = True

    @contract(d=str)
    def send_message(self, target, d):
        self.connection.privmsg(target, d)
#        print('to %s: %s' % (target, d))
        self.last_broadcast = time.time()


def start_irc(mi, server, channel):
    random.seed()

    ipfsi = IPFSInterface(mi.ipfs_path)
    identity = ipfsi.ipfs_id()
    hostname = socket.gethostname()
    printable = ''.join(ch for ch in hostname if ch.isalnum())
    printable = printable[:8]
    nickname = '%s_' % (__version__[-3:].replace('.', '')) + printable + '_' + identity[-4:]
    print('Using nickname %s' % nickname)

    delta_wait_for_welcome = 60
    stream_name = '/dns4/' + server[0] + '/tcp/' + str(server[1]) + '/irc'
    assert stream_name in mi.name2queue
    assert stream_name == mi.my_maddr

    key0 = CryptoKey(str(random.randint(0, 10000)), str(random.randint(0, 10000)))
    key0.hash = ipfsi.ipfs_id()

    t = time.time()
    irc_names = [
        stream_name + '/' + nickname,
        stream_name + '/' + channel,
    ]
    for irc_name in irc_names:
        msg = create_propose_message(('pri', identity, 'addresses'), irc_name, validity=[t, t + 300])
        mi.send_to_brain('', msg, sign=True)

    c = SwarmBot([server], stream_name, mi, channel, nickname)
    while True:

        while True:
            attempt_start = time.time()
            c.admitted = False
            print('server list: %s' % [_.host for _ in c.server_list])
            c._connect()
            while not c.admitted:
                # print('Waiting for welcome')
                time.sleep(1)
                c.reactor.process_once()

                delta = time.time() - attempt_start
                give_up = delta > delta_wait_for_welcome
                if give_up:
                    print('Could not receive welcome. giving up after %d s'
                          % delta_wait_for_welcome)
                    break
            if c.admitted:
                break
            print('Changing server')
            c.jump_server()

        c.disconnected = False

        while True:
            time.sleep(1.0)
            c.reactor.process_once(1.0)
            if c.disconnected:
                break

            xs = mi.get_many_for_me(timeout=1.0)

            for envelope in xs:
                try:
                    if not envelope.maddr_to or envelope.maddr_to[0] != '/':
                        msg = 'Invalid addr_to: %r' % envelope.maddr_to
                        print(msg)
                        continue

                    tokens = envelope.maddr_to.split('/')
                    assert tokens[0] == ''  # starts with '/'
                    dest = tokens[1]  # = '#duckiebots'
                    rest = '/' + "/".join(tokens[2:])

                    envelope.maddr_to = rest
                    envelope_json = envelope.to_json()

                    limit, chunk_size = 300, 150
                    try:
                        if len(envelope_json) < limit:
                            raw = [envelope_json]
                        else:
                            raw = fragmentize(envelope_json, max_chunk_length=chunk_size)

                        for r in raw:
                            if len(r) > 415:
                                print('Do not expect a string of len %s to be ok.' % len(r))
                            c.send_message(dest, r)
                    except MessageTooLong as e:
                        print(e)
                        msg = 'Envelope size is %s' % len(envelope_json)
                        print(msg)
#                    mi.send_to_brain(envelope.maddr_to, envelope.contents)
                except ServerNotConnectedError:
                    break




from Queue import Empty
from duckietown_swarm.brain import DistributedRepoSwarm
from duckietown_swarm.dcache_wire import create_request_message, MsgPong
from duckietown_swarm.ipfs_utils import IPFSInterface
from duckietown_swarm.packaging import find_more_information
import sys
import time

from contracts import contract
import yaml

from .dcache import Envelope, interpret_message
from .socket_utils import AsyncLineBasedConnection


def now_until(delta):
    t = time.time()
    return [t, t + delta]


@contract(returns=DistributedRepoSwarm)
def get_snapshot(host, port, what):

    albc = AsyncLineBasedConnection(host, port)
    maddr_from = '/cli'

    inbound = albc.get_q_inbound()
    outbound = albc.get_q_outbound()

    msg = create_request_message(what, validity=now_until(5))

    e = Envelope(maddr_from=maddr_from,
                 maddr_to='/dswarm/*/brain',
                 contents=msg)

    outbound.put(e)

    dc = DistributedRepoSwarm()
    from_channel = 'stdin'
    i = 0

#    wait_since_last = 15.0
#    t_last = time.time()
    while True:
        try:
            env = inbound.get(block=True, timeout=1.0)
#            t_last = time.time()
            sys.stderr.write('r')
        except Empty:
            pass
#            if i > 0:
#                if time.time() - t_last > wait_since_last:
#                    break
        else:
#            print env
            msg = interpret_message(env.contents)
            if isinstance(msg, MsgPong):
                break
            dc.process(env.contents, from_channel)
        i += 1
    print('processed %s messages' % i)
    return dc


def duckietown_swarm_cli_main():
    import argparse

    parser = argparse.ArgumentParser(description='Ping the swarm')
    parser.add_argument('host', type=str, default='localhost')
    parser.add_argument('what', default='*', help='what to request')

    args = parser.parse_args()

    host = args.host
    port = 8000
    what = args.what

    dc = get_snapshot(host, port, what)
    ipfsi = IPFSInterface(None)
    s = create_safe_index(ipfsi, dc)
    mh = ipfsi.add(s)
    print mh
    url = 'http://localhost:6080/ipfs/' + mh
    print url


def create_safe_index(ipfsi, dc):
    ''' Creates a webpage showing the safe data uploaded '''
    data = dc.summary_dict()

    print(yaml.dump(data))

    s = "<html><head></head><body>"

    verified = []

    all_mh = set(data['safe']) & set(data['verified']) & set(data['files'])

    find_provs = False

    for mh in all_mh:
#        s += '\n<a href="/ipfs/%s">%s</a>' % (v, v)
        try:
            more_info = find_more_information(ipfsi, mh, find_provs=find_provs)
#            print more_info
            sys.stderr.write(',')
        except Exception as e:
            print(e)
        else:
            verified.append(more_info)

#            if len(verified) > 10:
#                break
    verified = sorted(verified, key=lambda mi: mi.ctime)

    s += '''
        <style>
        .mh { font-family: monospace; }
        thead { font-weight: bold; }
        td { padding-left: 1em; }
        </style>

        <table>
        <thead><tr><td>Filename</td><td>Size</td><td>Upload by</td><td>node</td>
        <td></td>
        <td>safe</td>
        <td>verified</td>

        <td>ipfs_container</td>
        <td>ipfs_payload</td>
        <td>providers</td>

        </tr></thead>
        <tbody>
    '''
    for more_info in verified:
        d = more_info._asdict()
        d['is_safe'] = 'safe' if more_info.ipfs in data['safe'] else ''
        d['is_verified'] = 'verified' if more_info.ipfs in data['verified'] else ''
        d['date'] = more_info.ctime.strftime('%Y-%m-%d %H:%M')

        def shorten(mh):
            return '<a class="mh" href="/ipfs/%s">%s</a>' % (mh, mh[:4] + '...' + mh[-4:])

        def shorten_node(mh):
            return '<a class="mh" href="/ipns/%s">%s</a>' % (mh, mh[:4] + '...' + mh[-4:])

        d['upload_node_short'] = shorten_node(more_info.upload_node)
        d['ipfs_short'] = shorten(d['ipfs'])
        d['ipfs_payload_short'] = shorten(d['ipfs_payload'])

        d['providers'] = " ".join(map(shorten_node, more_info.providers_info))
        si = '''\n<tr><td><a href="/ipfs/{ipfs_payload}">{filename}</td><td>{size}</td>
        <td>{date}</td>
             <td>{upload_host}</td><td> {upload_node_short}</td>
             <td>{is_safe}</td>
             <td>{is_verified}</td>
              <td>{ipfs_short}</td>
              <td>{ipfs_payload_short}</td>
                <td>{providers}</td>
              </tr> '''

        s += si.format(**d)
#        ipfs='QmTJ6sFnB2Pxe1SG7hkd84gkv1YcaEai2tP86mbJKAdBdz', ipfs_payload='QmZYND3nJUW7KAmFMKNjHRMPmtL3FZtqSXW9qNRNcownoG', ipfs_info='QmbWEnQoAL8hPsk2GftZAYC4kTjK7HFYp6eM4D7ULyKA5r', filename='stasera.txt', upload_host='dorothy-7.local', upload_node='QmQDCfGCaYk7Uqc27k6gfgYD13pjny6MZSFQDk8Ra651Jo', upload_user='andrea', ctime=datetime.datetime(2018, 2, 7, 20, 49, 41), size=26, providers_payload=[], providers_info=[])

    s += '</tbody></table>'
    s += '</body>'
    s += '</html>'
    return s


if __name__ == '__main__':
    duckietown_swarm_cli_main()



from duckietown_swarm.dcache import CouldNotReadEnvelope
import socket
import time

from contracts.utils import raise_desc
from multiaddr import Multiaddr
import netifaces

from .dcache import Envelope
from .dcache_wire import create_propose_message

UDP_PORT = 1240


def get_suitable_udp_interfaces():
    ifaces = netifaces.interfaces()

    B = {}

    for iface in ifaces:

        # exclude docker interfaces
        if 'docker' in iface:
            continue
#        print('--- %s --- ' % iface)
        addrs = netifaces.ifaddresses(iface)
        if netifaces.AF_INET in addrs:
            ip4 = addrs[netifaces.AF_INET]
            for _ in ip4:
                if 'broadcast' in _:
                    B[iface] = {'address': str(_['addr']),
                                'broadcast': str(_['broadcast'])}

#        if netifaces.AF_LINK in addrs:
#            MAC = addrs[netifaces.AF_LINK]
#            print MAC

#    print('Broadcast addresses: %s' % B)
    return B


def get_mac_addresses():
    '''
        returns map ifname -> MAC
    '''
    ifaces = netifaces.interfaces()

    B = {}
    for iface in ifaces:
        addrs = netifaces.ifaddresses(iface)
        if netifaces.AF_LINK in addrs:
            MAC = addrs[netifaces.AF_LINK]
            B[iface] = MAC
    return B


def udp_broadcaster(mi):
    addr = mi.my_maddr
    m1 = Multiaddr(addr)

    # get the multiaddr protocol description objects
    p = m1.protocols()

    # [Protocol(code=4, name='ip4', size=32), Protocol(code=17, name='udp', size=16)]
    assert p[0].name == 'ip4'
    assert p[1].name == 'udp'
    # /ip4/IP/udp/port

    broadcast_address = m1.value_for_protocol(p[0].code)
    broadcast_port = int(m1.value_for_protocol(p[1].code))

    my_address = None
    interfaces = get_suitable_udp_interfaces()
    for name, x in interfaces.items():
        if x['broadcast'] == broadcast_address:
            my_address = x['address']
#            my_name = name
            print('Using interface %s = %s' % (name, x))

    if my_address is None:
        msg = 'Could not find address for broadcast address %s. (interface is down?)' % broadcast_address
        raise_desc(Exception, msg, interfaces=interfaces)

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    port = UDP_PORT
    while True:
        try:
            sock.bind((my_address, port))  # is add correct?
        except Exception as e:
            print('could not bind to %s:%s' % (my_address, port))
            print e
            port += 1
            if port > 2000:
                raise
            continue

        if port != UDP_PORT:
            print('finally bound to %s:%s' % (my_address, port))

#        print('ok for %s' % port)
        break

    while True:
        envelope = mi.get_next_for_me()
        envelope_json = envelope.to_json()
        sock.sendto(envelope_json, (broadcast_address, broadcast_port))


def udp_listener(mi):

    m1 = Multiaddr(mi.my_maddr)

    # get the multiaddr protocol description objects
    p = m1.protocols()

    # [Protocol(code=4, name='ip4', size=32), Protocol(code=17, name='udp', size=16)]
    assert p[0].name == 'ip4'
    assert p[1].name == 'udp'
    # /ip4/IP/udp/port

    broadcast_address = m1.value_for_protocol(p[0].code)
    broadcast_port = int(m1.value_for_protocol(p[1].code))

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    try:
        sock.bind((broadcast_address, broadcast_port))
    except Exception as e:
        msg = 'Could not bind UDP listener to %s:%s' % (broadcast_address, broadcast_port)
        msg += ' ' + str(e)
        raise Exception(msg)

    t = time.time()
    msg = create_propose_message(('pri', mi.key.hash, 'addresses'),
                                 mi.my_maddr, validity=[t, t + 300])
    mi.send_to_brain('', msg, sign=True)

    while True:
        msg, (_who_host, _who_port) = sock.recvfrom(4096)
        try:
            env = Envelope.from_json(msg)
        except CouldNotReadEnvelope as e:
            print(e)
            continue
        else:
            mi.dispatch(env)



from duckietown_swarm.brain import DistributedRepoSwarm
from duckietown_swarm.dcache import CryptoKey
from duckietown_swarm.dcache_wire import create_propose_message, create_dismiss_message
import time

from comptests.registrar import comptest, run_module_tests


@comptest
def init1():

    ds = DistributedRepoSwarm()
    key_admin = CryptoKey(None, 'mypublickey')
    key_trusted = CryptoKey('mysecretkeytrusted', 'mypublickey2')
    key_mine = CryptoKey('mykey', 'mykey11')

    ds.root.get_bucket('admin').propose(key_admin.hash, (time.time(), None), roles=['admin'])
    ds.root.get_bucket('trusted').propose(key_trusted.hash, (time.time(), None), roles=['admin'])
    msgs = []

    wm = create_propose_message('networks', '/pubsub/duckiebots')
    msgs.append(key_admin.generate_sign_message(wm))
    msgs.append(wm)

    wm = create_propose_message('networks', 'irc:frankfurt.co-design.science/#duckiebots',
                                validity=(0, time.time()))
    msgs.append(wm)

    wm = create_dismiss_message('networks', 'irc:frankfurt.co-design.science/#duckiebots',
                                validity=(time.time(), time.time() + 10))
    msgs.append(wm)

    for msg in msgs:
        ds.process(msg)

    print ds.summary()

    assert ds.query('networks') == ['/pubsub/duckiebots']

    # try to add an admin without a signature
    wm = create_propose_message('admin', key_mine.hash)
    try:
        ds.process(wm)
    except:
        pass
    else:
        msg = 'Invalid operation was accepted'
        raise Exception(msg)

#
#    ds.propose('networks',)
#    ds.propose('networks',)
#
#    ds.propose('files', 'QmPointerA')
#
#    entries = ds.bucket('networks')
#    assert entries[0].data == '/pubsub/duckiebots'
#    assert entries[0].validity.end == float('inf')
#    assert set(entries[0].channels) == set(['direct'])
#    c = entries[0].channels['direct']
#    assert isinstance(c, ChannelStats)

#    assert ChannelStats.counter == 1
#    ChannelStats.last
#    s.data  # data
#    s.validity  # t0, inf
#    s.channels  # channels this was received from
#    s.signatures  # collected signatures


if __name__ == '__main__':
    run_module_tests()



from duckietown_swarm.udp_interface import get_mac_addresses
from duckietown_swarm.utils import pretty_print_dictionary

if __name__ == '__main__':
    addresses = get_mac_addresses()
    print pretty_print_dictionary(addresses)


#! /usr/bin/env python

from collections import  namedtuple

from contracts.utils import raise_desc
from system_cmd import CmdException

from .utils import memoize_simple, yaml_load_omaps


class UnexpectedFormat(Exception):
    pass


MoreInfo = namedtuple('MoreInfo', 'ipfs ipfs_payload ipfs_info filename upload_host upload_node'
                      ' upload_user ctime size providers_payload providers_info')

INFO = 'upload_info1.yaml'


@memoize_simple
def find_more_information(ipfsi, h, find_provs=False):
    contents = ipfsi.ls(h, timeout="3s")
    timeout = "1h"
    if not INFO in contents:
        raise UnexpectedFormat(str(contents))

    try:
        info_data = ipfsi.get(h + '/' + INFO, timeout=timeout)
    except CmdException as e:
        if 'no link named' in e.res.stderr:
            raise UnexpectedFormat(e.res.stderr)
        raise

    try:
        info = yaml_load_omaps(info_data)
    except:
        raise UnexpectedFormat()

    if not isinstance(info, dict):
        msg = 'Expected a dict'
        raise_desc(UnexpectedFormat, msg, info=info)

    try:
        filename = info['filename']
        ctime = info['ctime']
        upload_host = info['upload_host']
        upload_node = info['upload_node']
        upload_user = info['upload_user']
    except KeyError as e:
        msg = 'Invalid format: %s' % e
        raise UnexpectedFormat(msg)

    try:
        ipfs_info = contents[INFO].hash
        ipfs_payload = contents[filename].hash
        size = contents[filename].size
    except KeyError as e:
        raise UnexpectedFormat(str(e))

    if find_provs:
        providers_info = ipfsi.dht_findprovs(ipfs_info)
        providers_payload = ipfsi.dht_findprovs(ipfs_payload)
    else:
        providers_info = []
        providers_payload = []

    return MoreInfo(ipfs=h,
                    ipfs_payload=ipfs_payload,
                    ipfs_info=ipfs_info,
                    filename=filename,
                    ctime=ctime,
                    size=size,
                    providers_info=providers_info,
                    providers_payload=providers_payload,
                    upload_host=upload_host,
                    upload_node=upload_node,
                    upload_user=upload_user)






import fnmatch
import socket
import time

from contracts import check_isinstance

from .dcache import DistributedRepo, ProcessFailure, interpret_message, Envelope
from .dcache_wire import MsgPing, create_pong_message, \
    MsgRequest, create_dismiss_message, create_propose_message, create_signature_message
from .ipfs_utils import IPFSInterface
from .irc2 import MessagingInterface
from .utils import DoAtInterval, get_sha256_base58


class DistributedRepoSwarm(DistributedRepo):

    def __init__(self):
        DistributedRepo.__init__(self)
        self.init_bucket('networks', allowed=[])
        self.init_bucket('peer', allowed=[])
        self.init_bucket('trusted', allowed=['trusted', 'admin'])
        self.init_bucket('admin', allowed=['admin'])

        self.init_bucket('files', allowed=[])
        self.init_bucket('verified', allowed=[])  #
        self.init_bucket('safe', allowed=[])  #

        self.init_bucket('summaries', allowed=[])  #

        self.init_bucket('pri', allowed=[])  #


MINUTES = 60
ALL_BRAINS = '/dswarm/*/brain'
DELTA_MAKE_SUMMARY = 5 * MINUTES
DAYS = 60 * 60 * 24


def brain(mi, cache_dir):
    check_isinstance(mi, MessagingInterface)

    ipfsi = IPFSInterface(mi.ipfs_path)

    dc = DistributedRepoSwarm()
    dc.use_cache_dir(cache_dir)

    t = time.time()

    proposed = set(dc.query('files'))
    verified = set(dc.query('verified'))
    for p in proposed:
        if not p in verified:
            mi.send_to_verifier(p)

    print('From cache: %s files' % len(dc.query('files')))
    print('From cache: %s verified' % len(dc.query('verified')))

    def on_new_files(d):
        data = d['data']
        #print('New file: %s' % data)
        mi.send_to_verifier(data)

    def on_new_peer(d):
        data = d['data']
        print('New peer: %s' % data)
        if data != identity:
            mi.send_to_connector(data)

    def on_new_verifed(d):
        data = d['data']
        #print('New verified: %s' % data)
        mi.send_to_pinner(data)

    def on_new_summary(d):
        bucket = d['bucket_name']
        data = d['data']
        if bucket[-1] == 'summaries':
            #print('New summary: %s' % d)
            which = bucket[1]
            if which != identity:
                print('Look up from %s %s' % (which, data))
                mi.send_to_readsummaries(data)
        else:
            pass
            #print('Ignoring new in bucket %s' % str(bucket))

    dc.add_hook_proposed('peer', on_new_peer)
    dc.add_hook_proposed('files', on_new_files)
    dc.add_hook_proposed('verified', on_new_verifed)
    dc.add_hook_proposed_all(on_new_summary)

    interval_make_summary = DoAtInterval(DELTA_MAKE_SUMMARY)
    DELTA_PEER = 10 * 60
    interval_advertise_peer = DoAtInterval(DELTA_PEER * 0.5)
    interval_addresses = DoAtInterval(10 * 60)
    interval_rebroadcast = DoAtInterval(10)

    key = mi.key
    identity = key.hash

    try:
        while True:
            if interval_addresses.its_time():
                addresses = ipfsi.get_addresses()
                for address in addresses:
                    if (address.startswith('/ip4/127.0.0.1/') or
                        address.startswith('/ip6/')):
                        continue
                    t = time.time()
                    msg = create_propose_message(('pri', identity, 'addresses'),
                                                 address, validity=[t, t + 300])
                    mi.send_to_brain('', msg, sign=True)

            if interval_make_summary.its_time():
                periodic_make_summary(dc, mi, identity, ipfsi)

            if interval_advertise_peer.its_time():
                t = time.time()
                msg = create_propose_message('peer', identity, validity=[t, t + DELTA_PEER])
                mi.send_to_brain('', msg, sign=True)

                import platform
                DELTA_INFO = 5 * DAYS
                properties = {}
                properties['hostname'] = socket.gethostname()
                properties['processor'] = platform.processor()
                properties['machine'] = platform.machine()
                properties['node'] = platform.node()
                properties['python_version'] = platform.python_version()
                properties['system'] = platform.system()

                for k, v in properties.items():
                    bucket = ('pri', identity, 'info', k)
                    msg = create_propose_message(bucket, v, validity=[t, t + DELTA_INFO])
                    mi.send_to_brain('', msg, sign=True)

            envelopes = mi.get_many_for_me(timeout=2)

            for envelope in envelopes:

                try:
                    m = interpret_message(envelope.contents)
                except:
                    pass
                else:
                    if isinstance(m, MsgPing):
                        reply_ping(mi, dc, envelope)

                    elif isinstance(m, MsgRequest):
                        reply_msg_request(mi, dc, envelope, m)

                    else:
                        try:
                            dc.process(envelope.contents, envelope.maddr_from)
                        except ProcessFailure as e:
                            em = 'Cannot interpret message: '
                            em += '\n' + str(e)
                            print(em)

            if interval_rebroadcast.its_time():
                periodic_rebroadcast(mi, dc)

            dc.cleanup()
            dc.sync()

    finally:
        print('Clean up cache')
        dc.close()


last_summary = None


def periodic_make_summary(dc, mi, identity, ipfsi):
    global last_summary
    summary_humans = dc.summary()
    summary_humans = summary_humans.replace(identity, 'QmSELF')
    summary_messages = dc.summary_messages()

    summary = ipfsi.get_tree_builder()
    summary.add_file_content('humans.txt', summary_humans)
    summary.add_file_content('machines.txt', '\n'.join(summary_messages))
    Q = summary.get_hash()

    bucket = ('pri', identity, 'summaries')
    t = time.time()
    if last_summary is not None:
        msg = create_dismiss_message(bucket, last_summary, validity=[t, None])
        mi.broadcast_to_all_brains(msg, sign=True)

    validity = [t, t + DELTA_MAKE_SUMMARY * 2]
    msg = create_propose_message(bucket, Q, validity)
    mi.broadcast_to_all_brains(msg, sign=True)

    last_summary = Q
    print('summary: %s' % last_summary)


def periodic_rebroadcast(mi, dc):
    messages = []

    for t, _ in dc.root.items():
        out_channels = mi._get_broadcasting_channels()
        these = dc.rebroadcast(t, out_channels)

        messages.extend(these)

    for envelope in messages:
        # find signatures
        hashed = get_sha256_base58(envelope.contents)
        if hashed in dc.signatures:
            for signer, signature in dc.signatures[hashed].items():
                msg = create_signature_message(hashed, signer, signature)
                e_sign = Envelope(envelope.maddr_from, envelope.maddr_to, msg)
                mi.dispatch(e_sign)
        mi.dispatch(envelope)


def reply_msg_request(mi, dc, envelope, m):
    messages = []
    pattern = m.bucket
    for bucket_name, _bucket in dc.root.items():
        c = "/" + "/".join(bucket_name)
        print('looking for %s' % str(bucket_name))
        if not fnmatch.fnmatch(c, pattern):
            print('does not match')
            continue
        out = dc._summary_message_for_bucket(bucket_name, only_positive=True)
#        print(bucket.summary())

        messages.extend(out)
    messages = [_[1] for _ in messages]
    messages = list(set(messages))

    print('Sending %d messages' % len(messages))
    for msg in messages:
        e = Envelope('', envelope.maddr_from, msg)
        mi.dispatch(e)

    pong = create_pong_message()
    e = Envelope('', envelope.maddr_from, pong)
    mi.dispatch(e)


def reply_ping(mi, _, envelope):
    pong = create_pong_message()
    e = Envelope('', envelope.maddr_from, pong)
    mi.dispatch(e)



from duckietown_swarm.dcache import create_propose_message, Envelope
from duckietown_swarm.ipfs_utils import IPFSInterface
import subprocess
import time

from system_cmd import system_cmd_result


def pubsub_reader_process(mi, topic):
    ipfsi = IPFSInterface(mi.ipfs_path)
    from duckietown_swarm.irc2 import CouldNotDispatch

    cmd = ['ipfs', 'pubsub', 'sub', '--discover', topic]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            env=ipfsi._get_env())

    # add my address
    maddr = '/ipfs-pubsub/' + topic
    t = time.time()
    msg = create_propose_message(('pri', mi.key.hash, 'addresses'),
                                 maddr, validity=[t, t + 300])
    mi.send_to_brain('', msg, sign=True)

    for line in iter(proc.stdout.readline, ''):
        line = line.strip()
        try:
            line = line.encode()
            env = Envelope.from_json(line)
#            e.maddr_from = '/ip4/%s/udp/%s' % (who_host, who_port) + e.maddr_from
#            mi.send_to_brain('', line)
        except ValueError as e:
            msg = 'Could not decode JSON: %s\n%r' % (e, line)
            print(msg)

        try:
            mi.dispatch(env)
        except CouldNotDispatch as e:
            print(e)


def pubsub_writer_process(mi, topic):
    ipfsi = IPFSInterface(mi.ipfs_path)
    while True:
        envelope = mi.get_next_for_me()
        envelope_json = envelope.to_json()
        if envelope.maddr_to == "":
            em = 'Will not write empty to: \n%s' % envelope.verbose()
            raise Exception(em)
#        print('writing: %s' % envelope)
        cmd = ['ipfs', 'pubsub', 'pub', topic, envelope_json + '\n']
        system_cmd_result('.', cmd, raise_on_error=False,
                          env=ipfsi._get_env())




from duckietown_swarm.fragments import fragmentize, Assembler
import random

from comptests.registrar import run_module_tests, comptest


@comptest
def test_chunks1():
    s = ''.join(str(i) for i in range(100))
    msgs = fragmentize(s, 50)
    print "\n".join(msgs)

    random.shuffle(msgs)
    a = Assembler()
    res = []
    for msg in msgs:
        a.push(msg)
        res.extend(a.pop())

    assert len(res) == 1
    print res


@comptest
def test_incomplete1():
    s = ''.join(str(i) for i in range(100))
    msgs = fragmentize(s, 50)
    print "\n".join(msgs)

    random.shuffle(msgs)
    a = Assembler()
    res = []
    for msg in msgs[:-1]:
        a.push(msg)
    res.extend(a.pop())

    assert len(res) == 0


@comptest
def test_corrupt():
    s = ''.join(str(i) for i in range(100))
    msgs = fragmentize(s, 50)
    print "\n".join(msgs)

    random.shuffle(msgs)
    a = Assembler()
    res = []
    for msg in msgs:
        msg = msg.replace('W', 'w')
        a.push(msg)
    res.extend(a.pop())

    assert len(res) == 0


if __name__ == '__main__':
    run_module_tests()
